/*
			(c) Copyright 1998-2000 - Tord Jansson
			======================================

		This file is part of the BladeEnc MP3 Encoder, based on
		ISO's reference code for MPEG Layer 3 compression, and might
		contain smaller or larger sections that are directly taken
		from ISO's reference code.

		All changes to the ISO reference code herein are either
		copyrighted by Tord Jansson (tord.jansson@swipnet.se)
		or sublicensed to Tord Jansson by a third party.

	BladeEnc is free software; you can redistribute this file
	and/or modify it under the terms of the GNU Lesser General Public
	License as published by the Free Software Foundation; either
	version 2.1 of the License, or (at your option) any later version.



	------------    Changes    ------------

	2000-11-10  Andr Piotrowski

	-	reformatted, used 'static' functions, global variables, less parameters

	2000-11-11  ap

	-	a lot of functions redesigned to let the bladetable become superfluous:
			genNoisePowTab(), iteration_loop(), calc_noise(), preemphasis(), amp_scalefac_bands()
	-	bladTabValue() replaced by cutting_crew()

	2000-11-22  ap

	-	bug fix:  module - reset fInit_huffman_read_flag

	---------------------------------------
*/

#define					OLD_CALC_SCFSI			1	/* There were lots of bugs, but let it on at the moment */
#define					BINARY_SEARCH			1	/* The binary search is faulty implemented, but let it on ... */



#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <assert.h>
#include "system.h"
#include "l3side.h"
#include "loop.h"
#include "huffman.h"
#include "l3bitstream.h"
#include "reservoir.h"
#include "loop-pvt.h"



double aint (double in);



/*
	Here are MPEG1 Table B.8 and MPEG2 Table B.1
	-- Layer III scalefactor bands.
	Index into this using a method such as:
	idx  = fr_ps->header->sampling_frequency
	       + (fr_ps->header->version * 3)
*/

struct scalefac_struct sfBandIndex[3] =
{
	{ /* Table B.8.b: 44.1 kHz */
		{0,4,8,12,16,20,24,30,36,44,52,62,74,90,110,134,162,196,238,288,342,418,576},
		{0,4,8,12,16,22,30,40,52,66,84,106,136,192}
	},
	{ /* Table B.8.c: 48 kHz */
		{0,4,8,12,16,20,24,30,36,42,50,60,72,88,106,128,156,190,230,276,330,384,576},
		{0,4,8,12,16,22,28,38,50,64,80,100,126,192}
	},
	{ /* Table B.8.a: 32 kHz */
		{0,4,8,12,16,20,24,30,36,44,54,66,82,102,126,156,194,240,296,364,448,550,576},
		{0,4,8,12,16,22,30,42,58,78,104,138,180,192}
	}
};



/*
	The following table is used to implement the scalefactor
	partitioning for MPEG2 as described in section
	2.4.3.2 of the IS. The indexing corresponds to the
	way the tables are presented in the IS:

	[table_number][row_in_table][column of nr_of_sfb]
*/

static unsigned nr_of_sfb_block[6][3][4] =
{
	{
		{ 6, 5, 5, 5},
		{ 9, 9, 9, 9},
		{ 6, 9, 9, 9}
	},
	{
		{6, 5,  7, 3},
		{9, 9, 12, 6},
		{6, 9, 12, 6}
	},
	{
		{11,10, 0, 0},
		{18,18, 0, 0},
		{15,18, 0, 0}
	},
	{
		{ 7, 7, 7, 0},
		{12,12,12, 0},
		{ 6,15,12, 0}
	},
	{
		{ 6, 6, 6, 3},
		{12, 9, 9, 6},
		{ 6,12, 9, 6}
	},
	{
		{ 8, 8, 5, 0},
		{15,12, 9, 0},
		{ 6,18, 9, 0}
	}
};



/* Table B.6: layer3 preemphasis */
int  pretab[21] =
{
	0, 0, 0, 0, 0, 0, 0,
	0, 0, 0, 0, 1, 1, 1, 
	1, 2, 2, 3, 3, 3, 2
};



/* This is the scfsi_band table from 2.4.2.7 of the IS */
int scfsi_band_long[5] = { 0, 6, 11, 16, 21 };



int						*scalefac_band_long  = &sfBandIndex[0].l[0];
int						*scalefac_band_short = &sfBandIndex[0].s[0];

int						fInit_iteration_loop;
int						fInit_huffman_read_flag = 0;

int						tjBitOverflow1;
int						tjBitOverflow2;



void fixStatic_loop( void )
{
	scalefac_band_long  = &sfBandIndex[0].l[0];
	scalefac_band_short = &sfBandIndex[0].s[0];
}





/*  ========================================================================  */
/*              generating the power tables                                   */
/*  ========================================================================  */



#define	BOT						200
#define POW216_MAX				(BOT/2 * 16)
#define POW216_MIN				(-25 * 16)



static	double					pow216_space[POW216_MAX-POW216_MIN+1];
static	double					*pow216 = pow216_space - POW216_MIN;
static	double					noisePowTab[8191+15];



void genNoisePowTab (void)
{
	int	i;

	for (i=POW216_MIN; i<=POW216_MAX; i++)
		pow216[i] = pow (2.0, (double)i/16.0);

	for (i=0; i<8191+15; i++)
		noisePowTab[i] = pow (i, 4.0/3.0);
}





/*  ========================================================================  */
/*              static variables                                              */
/*  ========================================================================  */



static	int						gr;   /* the current granule */
static	int						ch;   /* the current channel */



static	III_side_info_t			*side_info;   /* the current side information */
static	gr_info					*cod_info;    /* the current coding information */



static	double			   *xr_org_l;             /* the initial magnitudes of the spectral values */
static	double				  xr34_l[576];        /* the magnitudes powered by 3/4 */
static	int					   *ix_l;             /* quantized values */

static	double				energy_l[SFB_LMAX];
static	double				  xmin_l[SFB_LMAX];   /* the allowed distortion of the scalefactor band */
static	double				  xfsf_l[SFB_LMAX];   /* the current distortion of the scalefactor band */
static	int					expo16_l[SFB_LMAX];   /* sixteen times the scale factor band exponent */
static	int				 *scalefac_l;             /* the current scale factors */
static	int			   *scalefac_0_l;             /* scale factors for first granule */

static	double			  (*xr_org_s)[3];         /* some short block versions */
static	double				(*xr34_s)[3] = (double (*)[3]) xr34_l;
static	int					  (*ix_s)[3];

static	double				energy_s[SFB_SMAX][3];
static	double				  xmin_s[SFB_SMAX][3];
static	double				  xfsf_s[SFB_SMAX][3];
static	int					expo16_s[SFB_SMAX][3];
static	int				(*scalefac_s)[3];



static	int						max_used_sfb_l;
static	int						min_used_sfb_s;

static	int						end_sfb_l;
static	int						end_sfb_s;





/*  ========================================================================  */
/*              iteration_loop                                                */
/*  ========================================================================  */

void iteration_loop
(
	double					pe[][2],
	double					xr_org[2][2][576],
	III_psy_ratio			*ratio,
	III_side_info_t			*l3_side,
	int						l3_enc[2][2][576],
	int						mean_bits,
	int						stereo,
	double					xr_dec[2][2][576],
	III_scalefac_t			*scalefac,
	frame_params			*fr_ps,
	int						ancillary_pad,
	int						bitsPerFrame
)
{
	int						max_bits;
	int						i, sfb, b, scfsi_band;
	int						mode_gr;

	int						*main_data_begin;
	layer					*info;

	int						start, end;
	double					sum1, sum2, temp, x;



static double			en_tot[2][2];          /* gr, ch */
static double			en[2][2][SFB_LMAX-1];
static double			xm[2][2][SFB_LMAX-1];
static int				xrAllZero[2][2];

static double			log2;

#define					en_tot_krit				 10.0
#define					en_dif_krit				100.0
#define					en_scfsi_band_krit		 10.0
#define					xm_scfsi_band_krit		 10.0

	side_info = l3_side;

	main_data_begin = &side_info->main_data_begin;
	info = fr_ps->header;


	log2 = log(2.0);
	side_info->resvDrain = 0;

	if (!fInit_iteration_loop)
	{
		*main_data_begin = 0;
		fInit_iteration_loop = 1;
	}
	mode_gr = 2;

	scalefac_band_long  = &sfBandIndex[info->sampling_frequency].l[0];
	scalefac_band_short = &sfBandIndex[info->sampling_frequency].s[0];

	/* reading huffman code table */
	if (fInit_huffman_read_flag == 0) 
	{
		read_huffcodetab();
		fInit_huffman_read_flag = 1;
	}


	ResvFrameBegin (fr_ps, side_info, mean_bits, bitsPerFrame);

	for (gr=0; gr<mode_gr; gr++)
	{
		for (ch=0; ch<stereo; ch++)
		{
			xr_org_l = xr_org[gr][ch];
			xr_org_s = (double (*)[3]) xr_org_l;

			ix_l = l3_enc[gr][ch];
			ix_s = (int (*)[3]) ix_l;

			cod_info = &side_info->gr[gr].ch[ch].tt;

			scalefac_l = scalefac->l[gr][ch];  scalefac_0_l = scalefac->l[0][ch];
			scalefac_s = scalefac->s[gr][ch];

/*  ========    gr_deco    ========  */

	if (cod_info->window_switching_flag && (cod_info->block_type == 2))
	{
		if (cod_info->mixed_block_flag)
		{
			max_used_sfb_l = cod_info->sfb_lmax = 8;
			min_used_sfb_s = cod_info->sfb_smax = 3;
		}
		else
		{
			max_used_sfb_l = cod_info->sfb_lmax = 0;  /* No sb*/
			min_used_sfb_s = cod_info->sfb_smax = 0;
		}

		/* to access the entire array we need the last scalefac_band_short area */
		end_sfb_l = max_used_sfb_l /* cod_info->sfb_lmax */;
		end_sfb_s = SFB_SMAX;
	}
	else
	{
		max_used_sfb_l = cod_info->sfb_lmax = SFB_LMAX-1;
		min_used_sfb_s = cod_info->sfb_smax = SFB_SMAX-1;  /* No sb */

		/* to access the entire array we need the last scalefac_band_long area */
		end_sfb_l = SFB_LMAX;
		end_sfb_s = min_used_sfb_s /* cod_info->sfb_smax */;
	}


/*  ========    calc_xmin and start of quantanf_init    ========  */
/*
	Calculate the allowed distortion for each scalefactor band,
	as determined by the psychoacoustic model.
	xmin(sb) = ratio(sb) * energy(sb) / bandwidth(sb)
*/

	sum1 = 0.0;
	sum2 = 0.0;
	for (sfb=0; sfb<end_sfb_l; sfb++)
	{
		start = scalefac_band_long[sfb];
		end   = scalefac_band_long[sfb+1];
		expo16_l[sfb] = 0;
		temp = 0.0;
		for (i=start; i<end; i++)
		{
			if ((x = fabs(xr_org_l[i])) != 0.0)
			{
				xr34_l[i] = sqrt(x * sqrt(x));
				temp += x*x;
				sum1 += log(x);
			}
		}
		sum2 += energy_l[sfb] = temp;

		if (sfb < cod_info->sfb_lmax)
			xmin_l[sfb] = ratio->l[gr][ch][sfb] * temp / (end-start);
	}

	for (sfb=min_used_sfb_s; sfb<end_sfb_s; sfb++)
	{
		start = scalefac_band_short[sfb];
		end   = scalefac_band_short[sfb+1];
		for (b=0; b<3; b++)
		{
			expo16_s[sfb][b] = 0;
			temp = 0.0;
			for (i=start; i<end; i++)
			{
				if ((x = fabs(xr_org_s[i][b])) != 0.0)
				{
					xr34_s[i][b] = sqrt(x * sqrt(x));
					temp += x*x;
					sum1 += log(x);
				}
			}
			sum2 += energy_s[sfb][b] = temp;

			if (sfb < SFB_SMAX-1)
				xmin_s[sfb][b] = ratio->s[gr][ch][sfb][b] * temp / (end-start);
		}
	}


/*  ========    calc_scfsi    ========  */

#if OLD_CALC_SCFSI
	/*
		Formerly the maximum of |xr_l[gr][ch][i]| for i from 0 to 575 was assigned to an int variable xrmax[gr][ch].
		This maximum is always less than 1, so xrmax[gr][ch] was always set to 0!!!
	*/
	xrAllZero[gr][ch] = 1;
#else
	/* the total energy of the granule */
	temp = sum2;
	if (temp == 0.0)
		en_tot[gr][ch] = 0.0;
	else
		en_tot[gr][ch] = log(temp) / log2;

	xrAllZero[gr][ch] = (temp == 0.0);
#endif

	/* None of the granules contains short blocks */
	if (!cod_info->window_switching_flag || (cod_info->block_type != 2))
	{
		for (sfb=0; sfb<SFB_LMAX-1; sfb++)
		{
#if !OLD_CALC_SCFSI
			temp = energy_l[sfb];
			if (temp == 0.0)
				en[gr][ch][sfb] = 0.0;
			else
				en[gr][ch][sfb] = log(temp) / log2;
#endif
			if (xmin_l[sfb] == 0.0)
				xm[gr][ch][sfb] = 0.0;
			else
				xm[gr][ch][sfb] = log(xmin_l[sfb]) / log2;
		}

		if (gr == 1)
		{
			for (scfsi_band=0; scfsi_band<4; scfsi_band++)
				side_info->scfsi[ch][scfsi_band] = 0;

			/* The spectral values are not all zero */
			if (xrAllZero[0][ch] || xrAllZero[1][ch])  /* always, if OLD_CALC_SCFSI */
				goto calc_scfsi_break;

			if (fabs(en_tot[0][ch]-en_tot[1][ch]) >= en_tot_krit)
				goto calc_scfsi_break;

			temp = 0.0;
			for (sfb=0; sfb<SFB_LMAX-1; sfb++) 
				temp += fabs(en[0][ch][sfb]-en[1][ch][sfb]);
			if (temp >= en_dif_krit)
				goto calc_scfsi_break;

			for (scfsi_band=0; scfsi_band<4; scfsi_band++)
			{
				sum1 = sum2 = 0,0;

				start = scfsi_band_long[scfsi_band];
				end   = scfsi_band_long[scfsi_band+1];
				for (sfb=start; sfb<end; sfb++)
				{ 
					sum1 += fabs(en[0][ch][sfb]-en[1][ch][sfb]);
					sum2 += fabs(xm[0][ch][sfb]-xm[1][ch][sfb]);
				}

				if (sum1 < en_scfsi_band_krit && sum2 < xm_scfsi_band_krit)
					side_info->scfsi[ch][scfsi_band] = 1;
			}
		}
	}
	calc_scfsi_break:


			/* calculation of number of available bit( per granule ) */
			max_bits = ResvMaxBits (fr_ps, side_info, &pe[gr][ch], mean_bits);

			/* reset of iteration variables */

			for (sfb=0; sfb<SFB_LMAX-1; sfb++)
				scalefac_l[sfb] = 0;
			for (sfb=0; sfb<SFB_SMAX-1; sfb++)
				for (b=0; b<3; b++)
					scalefac_s[sfb][b] = 0;

			for (scfsi_band=0; scfsi_band<4; scfsi_band++)
				cod_info->slen[scfsi_band] = 0;

			cod_info->sfb_partition_table = &nr_of_sfb_block[0][0][0];
			cod_info->part2_3_length      = 0;
			cod_info->big_values          = 0;
			cod_info->count1              = 0;
			cod_info->scalefac_compress   = 0;
			cod_info->table_select[0]     = 0;
			cod_info->table_select[1]     = 0;
			cod_info->table_select[2]     = 0;
			cod_info->subblock_gain[0]    = 0;
			cod_info->subblock_gain[1]    = 0;
			cod_info->subblock_gain[2]    = 0;
			cod_info->region0_count       = 0;
			cod_info->region1_count       = 0;
			cod_info->part2_length        = 0;
			cod_info->preflag             = 0;
			cod_info->scalefac_scale      = 0;
			cod_info->quantizerStepSize   = 0.0;
			cod_info->count1table_select  = 0;

			/* all spectral values zero ? */
			if (sum2 != 0.0)
			{

/*  ========    quantanf_init (remaining)    ========  */

#define system_const			   8.0
#define minlimit				-100.0

	temp = nint (system_const * (sum1/288.0 - log(sum2/576.0)));
	if (temp < minlimit)
		temp = minlimit;
	/*
		SS 19-12-96. Starting value of
		global_gain or quantizerStepSize 
		has to be reduced for iteration_loop
	*/
	temp -= 70.0;


				cod_info->quantizerStepSize = temp;


				cod_info->part2_3_length = outer_loop (max_bits, fr_ps);
			}

			ResvAdjust (fr_ps, cod_info, side_info, mean_bits);

			cod_info->global_gain = nint (cod_info->quantizerStepSize + 210.0);
/*			assert (cod_info->global_gain < 256); */
		}	/* for ch */
	}	/* for gr */

	ResvFrameEnd (fr_ps, side_info, mean_bits);
}





/*  ========================================================================  */
/*              outer_loop                                                    */
/*  ========================================================================  */
/*
	The outer iteration loop controls the masking conditions
	of all scalefactorbands. It computes the best scalefac and
	global gain. This module calls the inner iteration loop
*/

static	int outer_loop
(
	int						max_bits,
	frame_params			*fr_ps
)
{
	int						scalesave_l[SFB_LMAX-1];
	int						scalesave_s[SFB_SMAX-1][3];
	int						bits, huff_bits, save_preflag, save_compress;
	int						sfb, b, over, iteration;

	iteration = 0;
	do 
	{
		iteration++;

		cod_info->part2_length = part2_length (fr_ps);
		huff_bits = max_bits - cod_info->part2_length;

#if BINARY_SEARCH   /* This function originally contains a bug!!!! */
		if (iteration == 1)
		{
			bin_search_StepSize (max_bits, cod_info->quantizerStepSize);  /* speeds things up a bit */
		}
#endif

		bits = inner_loop (huff_bits);

		calc_noise ();  /* distortion calculation */

		for (sfb=0; sfb<SFB_LMAX-1; sfb++)  /* save scaling factors */
			scalesave_l[sfb] = scalefac_l[sfb];

		for (sfb=0; sfb<SFB_SMAX-1; sfb++)
			for (b=0; b<3; b++)
				scalesave_s[sfb][b] = scalefac_s[sfb][b];

		save_preflag  = cod_info->preflag;
		save_compress = cod_info->scalefac_compress;

		preemphasis ();
		over = amp_scalefac_bands (iteration);

		if (loop_break ())
			break;
		if (scale_bitcount ())
			break;
	}
	while (over > 0);

	cod_info->preflag = save_preflag;
	cod_info->scalefac_compress = save_compress;

	for (sfb=0; sfb<SFB_LMAX-1; sfb++)
		scalefac_l[sfb] = scalesave_l[sfb];    

	for (sfb=0; sfb<SFB_SMAX-1; sfb++)
		for (b=0; b <3; b++)
			scalefac_s[sfb][b] = scalesave_s[sfb][b];

	cod_info->part2_length   = part2_length (fr_ps);
	cod_info->part2_3_length = cod_info->part2_length + bits;

	return cod_info->part2_3_length;
}





/*  ========================================================================  */
/*              inner_loop                                                    */
/*  ========================================================================  */
/*
	The code selects the best quantizerStepSize for a
	particular set of scalefacs
*/

static	int inner_loop
(
	int						max_bits
)
{
	int bits;

	cod_info->quantizerStepSize -= 1.0;
	do
	{
		do
		{
			cod_info->quantizerStepSize += 1.0;
			tjBitOverflow2 = FALSE;
			quantize ();
		}
		while (tjBitOverflow2);  /* within table range? */

		bits = count_bits ();
	}
	while (bits > max_bits);

	return bits;
}





/***************************************************************************/ 
/*        part2_length                                                     */ 
/***************************************************************************/ 

/* calculates the number of bits needed to encode the scalefacs in the     */
/* main data block                                                         */

static	int part2_length
(
	frame_params			*fr_ps
)
{
	int						slen1, slen2, bits;

	static	int				slen1_tab[16] = { 0, 0, 0, 0, 3, 1, 1, 1, 2, 2, 2, 3, 3, 3, 4, 4 };
	static	int				slen2_tab[16] = { 0, 1, 2, 3, 0, 1, 2, 3, 1, 2, 3, 1, 2, 3, 2, 3 };

	bits = 0;

	slen1 = slen1_tab[cod_info->scalefac_compress];
	slen2 = slen2_tab[cod_info->scalefac_compress];

	if (cod_info->window_switching_flag == 1 && cod_info->block_type == 2)
	{
		if (cod_info->mixed_block_flag)
			bits += (8+9) * slen1 + (18 * slen2);
		else
			bits += (18 * slen1) + (18 * slen2);
	}
	else if (gr==0)
		bits += (6+5) * slen1 + (5+5) * slen2;
	else
	{
		if (side_info->scfsi[ch][0] == 0)  bits += (6 * slen1);
		if (side_info->scfsi[ch][1] == 0)  bits += (5 * slen1);
		if (side_info->scfsi[ch][2] == 0)  bits += (5 * slen2);
		if (side_info->scfsi[ch][3] == 0)  bits += (5 * slen2);
	}

	return bits;
}





/*************************************************************************/
/*            scale_bitcount                                             */
/*************************************************************************/


/* Also calculates the number of bits necessary to code the scalefactors. */

static	int scale_bitcount (void)
{
    int i, k, sfb, max_slen1 = 0, max_slen2 = 0, /*a, b, */ ep = 2;


    static int slen1[16] = { 0, 0, 0, 0, 3, 1, 1, 1, 2, 2, 2, 3, 3, 3, 4, 4 };
    static int slen2[16] = { 0, 1, 2, 3, 0, 1, 2, 3, 1, 2, 3, 1, 2, 3, 2, 3 };
    static int pow2[5]   = { 1, 2, 4, 8, 16 };

    if (cod_info->window_switching_flag != 0 && cod_info->block_type == 2)
    {
        if (cod_info->mixed_block_flag == 0)
        {
            /* a = 18; b = 18;  */
            for (i=0; i<3; i++)
            {
                for (sfb=0; sfb<6; sfb++)
                    if (scalefac_s[sfb][i] > max_slen1)
                        max_slen1 = scalefac_s[sfb][i];
                for (sfb=6; sfb<12; sfb++)
                    if (scalefac_s[sfb][i] > max_slen2)
                        max_slen2 = scalefac_s[sfb][i];
            }
        }
        else
        {/* mixed_block_flag = 1 */
            /* a = 17; b = 18;  */
            for (sfb=0; sfb<8; sfb++)
                if (scalefac_l[sfb] > max_slen1)
                    max_slen1 = scalefac_l[sfb];
            for (i=0; i<3; i++)
            {
                for (sfb=3; sfb<6; sfb++)
                    if (scalefac_s[sfb][i] > max_slen1)
                        max_slen1 = scalefac_s[sfb][i];
                for (sfb=6; sfb<12; sfb++)
                    if (scalefac_s[sfb][i] > max_slen2)
                        max_slen2 = scalefac_s[sfb][i];
            }
        }

    }
    else
    { /* block_type == 1,2,or 3 */
        /* a = 11; b = 10;   */
        for (sfb=0; sfb<11; sfb++)
            if (scalefac_l[sfb] > max_slen1)
                max_slen1 = scalefac_l[sfb];
        for (sfb=11; sfb<21; sfb++)
            if (scalefac_l[sfb] > max_slen2)
                max_slen2 = scalefac_l[sfb];
    }

    for (k=0; k<16; k++)
    {
        if (max_slen1 < pow2[slen1[k]] && max_slen2 < pow2[slen2[k]])
        { 
            ep = 0;
            break;
        } 
    }

    if (ep == 0)
        cod_info->scalefac_compress = k;

    return ep;
}





/*************************************************************************/
/*            calc_noise                                                 */
/*************************************************************************/

/*   Function: Calculate the distortion introduced by the quantization   */
/*   in each scale factor band.                                          */

static	void calc_noise (void)
{
	int						i, b, sfb, start, end;
	double					f, g, sum, temp;


	f = pow216[-4 * (int)cod_info->quantizerStepSize];

	for (sfb=0; sfb<max_used_sfb_l; sfb++)
	{
			start = scalefac_band_long[sfb];
			end   = scalefac_band_long[sfb+1];

			sum = 0.0;

			g = pow216[expo16_l[sfb]];

			for (i=start; i<end; i++)
			{
				temp = fabs(xr_org_l[i]) * g - noisePowTab[ix_l[i]] / f;
				sum += temp * temp;
			}

			xfsf_l[sfb] = sum / (end-start);
	}

	for (b=0; b<3; b++)
	{
		f = pow216[-4 * ((int)cod_info->quantizerStepSize + 8 * cod_info->subblock_gain[b])];

		for (sfb=min_used_sfb_s; sfb<SFB_SMAX-1; sfb++)
		{
				start = scalefac_band_short[sfb];
				end   = scalefac_band_short[sfb+1];

				sum = 0.0;

				g = pow216[expo16_s[sfb][b]];

				for (i=start; i<end; i++)
				{
	 				temp = fabs(xr_org_s[i][b]) * g - noisePowTab[ix_s[i][b]] / f;
					sum += temp * temp;
				}       

				xfsf_s[sfb][b] = sum / (end-start);
		}
	}
}





/*************************************************************************/
/*            loop_break                                                 */
/*************************************************************************/

/*  Function: Returns zero if there is a scalefac which has not been
    amplified. Otherwise it returns one. 
*/

static	int loop_break (void)
{
    uint					i, sfb;

    for (sfb=0; sfb<cod_info->sfb_lmax; sfb++)
        if (scalefac_l[sfb] == 0)
            return 0;

    for (sfb=cod_info->sfb_smax; sfb<12; sfb++)
        for (i=0; i<3; i++)
            if (scalefac_s[sfb][i] == 0)
                return 0;

    return 1;
}





/*************************************************************************/
/*            preemphasis                                                */
/*************************************************************************/

/*
  See ISO 11172-3  section  C.1.5.4.3.4
*/

static	void preemphasis (void)
{
	int						sfb, scfsi_band, over;
	int						expo16_off;

	if (gr == 1)
	{
		/*
			If the second granule is being coded and scfsi is active in
			at least one scfsi_band, the preemphasis in the second granule
			is set equal to the setting in the first granule
		*/
		for ( scfsi_band = 0; scfsi_band < 4; scfsi_band++ )
			if ( side_info->scfsi[ch][scfsi_band] )
			{
				cod_info->preflag = side_info->gr[0].ch[ch].tt.preflag;
				return;
			}
	}

	/*
		Preemphasis is switched on if in all the upper four scalefactor
		bands the actual distortion exceeds the threshold after the
		first call of the inner loop
	*/
	if (cod_info->block_type != 2 && cod_info->preflag == 0)
	{	
		over = 0;
		for (sfb=17; sfb<21; sfb++)
			if (xfsf_l[sfb] > xmin_l[sfb])
				over++;

		if (over == 4)
		{
			cod_info->preflag = 1;

			expo16_off = 16 * (1 + cod_info->scalefac_scale) / 2;

			for (sfb=0; sfb<cod_info->sfb_lmax; sfb++)
			{
				expo16_l[sfb] += expo16_off * pretab[sfb];
				xmin_l[sfb] *= pow216[2 * expo16_off * pretab[sfb]];

			}
		}
	}
}





/*************************************************************************/
/*            amp_scalefac_bands                                         */
/*************************************************************************/

/* 
  Amplify the scalefactor bands that violate the masking threshold.
  See ISO 11172-3 Section C.1.5.4.3.5
*/

static	int amp_scalefac_bands
(
	int						iteration
)
{
	int						sfb, b, scfsi_band, over = 0;
	int						copySF, preventSF;
	gr_info					*gr0;
	int						expo16_off;

	gr0 = &side_info->gr[0].ch[ch].tt;

	copySF = 0;
	preventSF = 0;

	expo16_off = 16 * (1 + cod_info->scalefac_scale) / 2;

	if (gr == 1)
	{
		for (scfsi_band=0; scfsi_band<4; scfsi_band++)
			if (side_info->scfsi[ch][scfsi_band])
			{
				expo16_off = 16 * (1 + gr0->scalefac_scale) / 2;

				if (iteration == 1)
					copySF = 1;
				else
					preventSF = 1;

				break;
			}
	}

	scfsi_band = 0;

	for (sfb=0; sfb<cod_info->sfb_lmax; sfb++)
	{
		if (copySF || preventSF)
		{
			if ((int) sfb == scfsi_band_long[scfsi_band + 1])
				scfsi_band += 1;
			if (side_info->scfsi[ch][scfsi_band])
			{
				if (copySF)
					scalefac_l[sfb] = scalefac_0_l[sfb];
				continue;
			}
		}	    
		if (xfsf_l[sfb] > xmin_l[sfb])
		{
			over++;

			expo16_l[sfb] += expo16_off;
			xmin_l[sfb] *= pow216[2 * expo16_off];

			scalefac_l[sfb]++;
		}
	}

	for (b=0; b<3; b++)
		for (sfb=cod_info->sfb_smax; sfb<12; sfb++)
			if (xfsf_s[sfb][b] > xmin_s[sfb][b])
			{
				over++;

				expo16_s[sfb][b] += expo16_off;
				xmin_s[sfb][b] *= pow216[2 * expo16_off];

				scalefac_s[sfb][b]++;
			}

	return over;
}





/*  ========================================================================  */
/*              quantize                                                      */
/*  ========================================================================  */
/*
	Quantization of the vector xr ( -> ix)
*/

int INLINE cutting_crew (FLOAT in)
{
	int						retVal;

	retVal = (int) (in + 0.4054);
	if (retVal > 8192)
	{
		tjBitOverflow1 = TRUE;
		if (retVal > 8191+14)
			tjBitOverflow2 = TRUE;
	}

	return retVal;
}



static	void quantize (void)
{
	int						sfb, i, b, start, end;
	double					f, z, y;

	for (sfb=0; sfb<end_sfb_l; sfb++)
	{
		start = scalefac_band_long[sfb];
		end   = scalefac_band_long[sfb+1];

			        /* (expo16_l[sfb] - 16/4 * quant_step) * 3/4 */
			f = pow216[(expo16_l[sfb]/4 - (int)cod_info->quantizerStepSize) * 3];

			for (i=start; i<end; i+=2)
			{
				z = xr34_l[i  ] * f;
				y = xr34_l[i+1] * f;
				ix_l[i  ] = cutting_crew (z);
				ix_l[i+1] = cutting_crew (y);
			}
	}

	for (sfb=min_used_sfb_s; sfb<end_sfb_s; sfb++)
	{
		start = scalefac_band_short[sfb];
		end   = scalefac_band_short[sfb+1];

		for (b=0; b<3; b++)
		{
				        /* (expo_s[sfb][b] - 16/4 * (quant_step + 8 * cod_info->subblock_gain[b])) * 3/4 */
				f = pow216[(expo16_s[sfb][b] / 4 - (int)cod_info->quantizerStepSize - 8 * cod_info->subblock_gain[b]) * 3];

				for (i=start; i<end; i+=2)
				{
					z = xr34_s[i  ][b] * f;
					y = xr34_s[i+1][b] * f;
					ix_s[i  ][b] = cutting_crew (z);
					ix_s[i+1][b] = cutting_crew (y);
				}
		}
	}
}





/*  ========================================================================  */
/*              ix_max_tj                                                     */
/*  ========================================================================  */
/*
	Calculate the maximum of ix_l from begin to end
*/


static	int ix_max_tj
(
	unsigned int			begin,
	unsigned int			end
)
{
	int						i;
	int						max = 0;

	for (i=begin; i<end; i++)
		if (ix_l[i] > max)
			max = ix_l[i];

	return max;
}





/*  ========================================================================  */
/*              count_bits                                                    */
/*  ========================================================================  */

struct
{
	unsigned region0_count;
	unsigned region1_count;
} subdv_table[ 23 ] =
{
	{0, 0}, /*  0 bands */
	{0, 0}, /*  1 bands */
	{0, 0}, /*  2 bands */
	{0, 0}, /*  3 bands */
	{0, 0}, /*  4 bands */
	{0, 1}, /*  5 bands */
	{1, 1}, /*  6 bands */
	{1, 1}, /*  7 bands */
	{1, 2}, /*  8 bands */
	{2, 2}, /*  9 bands */
	{2, 3}, /* 10 bands */
	{2, 3}, /* 11 bands */
	{3, 4}, /* 12 bands */
	{3, 4}, /* 13 bands */
	{3, 4}, /* 14 bands */
	{4, 5}, /* 15 bands */
	{4, 5}, /* 16 bands */
	{4, 6}, /* 17 bands */
	{5, 6}, /* 18 bands */
	{5, 6}, /* 19 bands */
	{5, 7}, /* 20 bands */
	{6, 7}, /* 21 bands */
	{6, 7}, /* 22 bands */
};



/*
	Calculation of rzero, count1, big_values
	(Partitions ix into big values, quadruples and zeros).

	Determines the number of bits to encode the quadruples.

	Presumable subdivides the bigvalue region which will
	use separate Huffman tables.

	Select huffman code tables for bigvalues regions

	Count the number of bits necessary to code the bigvalues region.
*/

static	int count_bits (void)
{
	int						zero_region = 576;
	int						bigv_region = 576;

	int						bits = 0;
	int						sum0 = 0;
	int						sum1 = 0;

	if (!cod_info->window_switching_flag || (cod_info->block_type != 2))
	{
		int						p;

		for (; zero_region>1; zero_region-= 2)
			     if (ix_l[zero_region-1])  break;
			else if (ix_l[zero_region-2])  break;

		for (bigv_region=zero_region; bigv_region>3; bigv_region-= 4)
		{
			     if (ix_l[bigv_region-1] > 1)  break;
			else if (ix_l[bigv_region-2] > 1)  break;
			else if (ix_l[bigv_region-3] > 1)  break;
			else if (ix_l[bigv_region-4] > 1)  break;

			p = 0;
			if (ix_l[bigv_region-1])  bits++, p |= 8;
			if (ix_l[bigv_region-2])  bits++, p |= 4;
			if (ix_l[bigv_region-3])  bits++, p |= 2;
			if (ix_l[bigv_region-4])  bits++, p |= 1;

			sum0 += ht[32].hlen[p];
			sum1 += ht[33].hlen[p];
		}
	}
	cod_info->count1     = (zero_region-bigv_region) / 4;
	cod_info->big_values =              bigv_region  / 2;

	if (sum0 < sum1)
	{
		bits += sum0;
		cod_info->count1table_select = 0;
	}
	else
	{
		bits += sum1;
		cod_info->count1table_select = 1;
	}

	if (bigv_region)
	{
		if (cod_info->window_switching_flag)
		{
			if (cod_info->mixed_block_flag || (cod_info->block_type != 2))
			{
				cod_info->region0_count =  7;
				cod_info->region1_count = 13;
				cod_info->address1 = scalefac_band_long[cod_info->region0_count+1];
			}
			else
			{ 
				cod_info->region0_count =  8;
				cod_info->region1_count = 36;
				cod_info->address1 = 36;
			}
			cod_info->address2 = bigv_region;
			cod_info->address3 = 0;
		}
		else
		{	/* long blocks */

/*  ========    This area replaced my favourite joke in original subdivide()    ========  */

			int						index0, index1;
			int						scfb_anz = 0;

			/* Calculate scfb_anz */
			while (scalefac_band_long[scfb_anz] < bigv_region)
				scfb_anz++;
/*			assert (scfb_anz < 23); */

			index0 = (cod_info->region0_count = subdv_table[scfb_anz].region0_count) + 1;
			index1 = (cod_info->region1_count = subdv_table[scfb_anz].region1_count) + 1;

			cod_info->address1 = scalefac_band_long[index0];
			cod_info->address2 = scalefac_band_long[index0 + index1];
			cod_info->address3 = bigv_region;

/*  ======================================================================================  */

		}
	}
	else
	{	/* no big_values region */
		cod_info->region0_count = 0;
		cod_info->region1_count = 0;
	}


	cod_info->table_select[0] = 0;
	cod_info->table_select[1] = 0;
	cod_info->table_select[2] = 0;

	if (cod_info->window_switching_flag && (cod_info->block_type == 2))
	{
		/*
			Within each scalefactor band, data is given for successive
			time windows, beginning with window 0 and ending with window 2.
			Within each window, the quantized values are then arranged in
			order of increasing frequency...
		*/
		int						sfb, window, line, start, end;
		unsigned int			max1, max2, *pmax, tMax;
		int						x, y;

		max1 = max2 = 0;

		for (sfb=0; sfb<SFB_SMAX; sfb++)
		{
			start = scalefac_band_short[sfb];
			end   = scalefac_band_short[sfb+1];

			if (start < 12)
				pmax = &max1;
			else
				pmax = &max2;

			tMax = *pmax;

			for (window=0; window<3; window++)
			{
				for (line=start; line<end; line+=2)
				{
					x = ix_s[line  ][window];
					y = ix_s[line+1][window];

					if (x > tMax)  tMax = x;
					if (y > tMax)  tMax = y;
				}
			}

			*pmax = tMax;
		}
		cod_info->table_select[0] = choose_table(max1);
		cod_info->table_select[1] = choose_table(max2);

		if (cod_info->mixed_block_flag)
		{
			if (cod_info->table_select[0])
				bits += count_bit_tj (0, cod_info->address1, cod_info->table_select[0]);
			sfb = 2;
		}
		else
			sfb = 0;

		for ( ; sfb<SFB_SMAX; sfb++)
		{
			unsigned				tableindex;
			unsigned				idx;
			struct huffcodetab		*h;

			start = scalefac_band_short[sfb];
			end   = scalefac_band_short[sfb+1];

			if (start < 12)
				tableindex = cod_info->table_select[0];
			else
				tableindex = cod_info->table_select[1];
/*			assert (tableindex < 32); */

			if (tableindex != 0)
			{
				h = &(ht[tableindex]);

				for (window=0; window<3; window++)
				{
					for (line=start; line<end; line+=2)
					{
						int x = ix_s[line  ][window];
						int y = ix_s[line+1][window];

						if (tableindex > 15)
						{	/* ESC-table is used */
							if (x > 14)
							{
								x = 15;
								bits += h->linbits;
							}
							if (y > 14)
							{
								y = 15;
								bits += h->linbits;
							}
						}

						idx = (x * h->ylen) + y;
						bits += h->hlen[ idx ];

						if (x)  bits++;
						if (y)  bits++;
					}
				}
			}
		}
	}
	else
	{
		if (cod_info->address1 > 0)
			if ((cod_info->table_select[0] = new_choose_table_tj (0, cod_info->address1)) > 0)
				bits += count_bit_tj (0, cod_info->address1, cod_info->table_select[0]);  /* region0 */

		if (cod_info->address2 > cod_info->address1)
			if ((cod_info->table_select[1] = new_choose_table_tj (cod_info->address1, cod_info->address2)) > 0)
				bits += count_bit_tj (cod_info->address1, cod_info->address2, cod_info->table_select[1]);  /* region1 */

		if (cod_info->big_values * 2 > cod_info->address2)
			if ((cod_info->table_select[2] = new_choose_table_tj (cod_info->address2, cod_info->big_values*2)) > 0)
				bits += count_bit_tj (cod_info->address2, cod_info->address3, cod_info->table_select[2]);  /* region2 */
	}

	return bits;
}





/*  ========================================================================  */
/*              new_choose table                                              */
/*  ========================================================================  */
/*
	Choose the Huffman table that will encode ix[begin..end] with
	the fewest bits.

	Note: This code contains knowledge about the sizes and characteristics
	of the Huffman tables as defined in the IS (Table B.7), and will not work
	with any arbitrary tables.
*/

static	int new_choose_table_tj
(
	unsigned int			begin,
	unsigned int			end
)
{
	uint					max, choice0, choice1;
	int						sum0, sum1;

	max = ix_max_tj (begin, end);

	if (max == 0)
		return 0;

	choice0 = 0;
	choice1 = 0;

	if (max < 15)
	{
		choice0 = 1;  /* not 0 -- ht[0].xlen == 0 */
		while (ht[choice0].xlen <= max)
			choice0++;

		sum0 = count_bit_tj (begin, end, choice0);

		switch ( choice0 )
		{
			case  2:  sum1 = count_bit_tj (begin, end,  3); if (sum1 <= sum0) choice0 =  3; break;
			case  5:  sum1 = count_bit_tj (begin, end,  6); if (sum1 <= sum0) choice0 =  6; break;

			case  7:  sum1 = count_bit_tj (begin, end,  8); if (sum1 <= sum0)  {choice0 =  8; sum0 = sum1;}
					  sum1 = count_bit_tj (begin, end,  9); if (sum1 <= sum0) choice0 =  9; break;

			case 10:  sum1 = count_bit_tj (begin, end, 11); if (sum1 <= sum0)  {choice0 = 11; sum0 = sum1;}
					  sum1 = count_bit_tj (begin, end, 12); if (sum1 <= sum0) choice0 = 12; break;

			case 13:  sum1 = count_bit_tj (begin, end, 15); if (sum1 <= sum0) choice0 = 15; break;

			default:  break;
		}
	}
	else
	{
		/* try tables with linbits */
		max -= 15;

		choice0 = 15;
		while (ht[choice0].linmax < max)
			choice0++;

		choice1 = 24;
		while (ht[choice1].linmax < max)
			choice1++;

		sum0 = count_bit_tj (begin, end, choice0);
		sum1 = count_bit_tj (begin, end, choice1);
		if (sum1 < sum0)
			choice0 = choice1;
	}

	return choice0;
}





/*  ========================================================================  */
/*              choose table                                                  */
/*  ========================================================================  */

static	int choose_table (unsigned int max)
{
	unsigned int			choice0;

	if (max == 0)
		return 0;

	if ( max < 15 )
	{
		choice0 = 1;  /* not 0 -- ht[0].xlen == 0 */
		while (ht[choice0].xlen <= max)
			choice0++;
	}
	else
	{	
		max -= 15;

		choice0 = 15;
		while (ht[choice0].linmax < max)
			choice0++;
	}

	return choice0;
}





/*  ========================================================================  */
/*              count_bit                                                     */
/*  ========================================================================  */
/*
	Count the number of bits necessary to code the subregion. 
*/

static	int count_bit_tj
(
	unsigned int			start,
	unsigned int			end,
	unsigned int			table
)
{
	int						i, sum;
	unsigned				idx;
	struct huffcodetab		*h;
	int						x, y;

	if (table == 0)
		return 0;

	h = &(ht[table]);

	sum = 0;
	for (i=start; i<end; i+=2)
	{
		x = ix_l[i];
		y = ix_l[i+1];

		if (table > 15)
		{
			if (x > 14)
			{
				x = 15;
				sum += h->linbits;
			}
			if (y > 14)
			{
				y = 15;
				sum += h->linbits;
			}
		}

		idx = (x * h->ylen) + y;
		sum += h->hlen[idx];

		if (x)  sum++;
		if (y)  sum++;
	}

	return sum;
}





#ifndef HAVE_NINT
int nint (double in)
{
	int						temp;

	if (in < 0)
		temp = (int)(in - 0.5);
	else
		temp = (int)(in + 0.5);

	return temp;
}

double aint (double in)
{
	return((int) in);
}
#endif





/*  ========================================================================  */
/*              bin_search_step_size                                          */
/*  ========================================================================  */
/*
	The following optional code written by Seymour Shlien
	will speed up the outer_loop code which is called
	by iteration_loop. When BIN_SEARCH is defined, the
	outer_loop function precedes the call to the function inner_loop
	with a call to bin_search gain defined below, which
	returns a good starting quantizerStepSize.

	The function count_bits() [a sequence of statements, originally part of inner_loop()]
	was completely rewritten.
*/

static	int bin_search_StepSize
(
	int						desired_rate,
	double					start
)
{
	double					top, bot, next, last;
	int						bit;

	top = start;
	bot = 200;

	next = start;
	do
	{
		last = next;
		next = aint((top+bot)/2.0);
		cod_info->quantizerStepSize = next;

		tjBitOverflow1 = FALSE;
		quantize ();
		if (tjBitOverflow1)
			bit = 100000;
		else
			bit = count_bits ();

		if (bit>desired_rate) 
			top = next;
		else 
			bot = next;
	}
	while ((bit != desired_rate) && (fabs(last-next) > 1.0));

	return next;
}



